// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.notification.provider.azure;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnitRunner;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.notification.provider.azure.pubsub.EventGridRequestBodyExtractor;
import org.springframework.http.HttpStatus;

import javax.servlet.http.HttpServletRequest;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.Map;

import static org.junit.Assert.fail;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;

@RunWith(MockitoJUnitRunner.class)
public class EventGridRequestBodyExtractorTest {

    private EventGridRequestBodyExtractor sut;

    @Mock
    private HttpServletRequest httpServletRequest;

    @Mock
    private JaxRsDpsLog log;

    @Test
    public void should_returnTrue_isHandshakeRequest() throws IOException {
        // Set up
        String validHandshakeRequestRoot =
                "    [{\n" +
                        "        \"id\": \"testId\",\n" +
                        "        \"topic\": \"testTopic\",\n" +
                        "        \"subject\": \"\",\n" +
                        "        \"data\": {\n" +
                        "            \"validationCode\": \"testValidationCode\",\n" +
                        "            \"validationUrl\": \"testURL\"\n" +
                        "        },\n" +
                        "        \"eventType\": \"Microsoft.EventGrid.SubscriptionValidationEvent\",\n" +
                        "        \"eventTime\": \"2020-08-14T11:18:55.9278057Z\",\n" +
                        "        \"metadataVersion\": \"1\",\n" +
                        "        \"dataVersion\": \"2\"\n" +
                        "    }]";
        BufferedReader reader = new BufferedReader(new StringReader(validHandshakeRequestRoot));
        when(this.httpServletRequest.getReader()).thenReturn(reader);
        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

        // Act
        boolean response = this.sut.isHandshakeRequest();

        // Assert
        Assert.assertTrue(response);
    }

    @Test
    public void shouldThrow_whenRequestTypeIsNotright_isHandshakeRequest() throws IOException {
        //SetUp
        String invalidHandshakeRequestRoot =
                "    [{\n" +
                        "        \"id\": \"testId\",\n" +
                        "        \"topic\": \"testTopic\",\n" +
                        "        \"subject\": \"\",\n" +
                        "        \"data\": {\n" +
                        "            \"validationCode\": \"testValidationCode\",\n" +
                        "            \"validationUrl\": \"testURL\"\n" +
                        "        },\n" +
                        "        \"eventType\": \"SubscriptionValidationEvent\",\n" +
                        "        \"eventTime\": \"2020-08-14T11:18:55.9278057Z\",\n" +
                        "        \"metadataVersion\": \"1\",\n" +
                        "        \"dataVersion\": \"2\"\n" +
                        "    }]";
        BufferedReader reader = new BufferedReader(new StringReader(invalidHandshakeRequestRoot));
        when(this.httpServletRequest.getReader()).thenReturn(reader);

        try{
            // Act
            sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

            // Assert
            fail("Should Throw Exception");
        } catch (AppException appException){
            Assert.assertEquals(HttpStatus.BAD_REQUEST.value(), appException.getError().getCode());
            Assert.assertEquals("Unable to parse request payload.", appException.getError().getMessage());
        } catch (Exception exception) {
            fail("Should Throw AppException");
        }
    }

    @Test
    public void should_throwWhenAttributesAreMissing_extractDataFromRequestBody() throws IOException {
        String requestRootWithoutAttributes = "[{\n" +
                "        \"id\": \"2425\",\n" +
                "        \"eventType\": \"recordInserted\",\n" +
                "        \"subject\": \"myapp/vehicles/motorcycles\",\n" +
                "        \"data\": {\n" +
                "            \"data\": \"dGVzdA==\",\n" +
                "            \"messageId\": \"136969346945\"\n" +
                "        },\n" +
                "        \"dataVersion\": \"1.0\",\n" +
                "        \"metadataVersion\": \"1\",\n" +
                "        \"eventTime\": \"2020-08-14T18:04:12+00:00\",\n" +
                "        \"topic\": \"/subscriptions/asdf/resourceGroups/komakkar-OSDU-RG/providers/Microsoft.EventGrid/topics/recordChanged\"\n" +
                "    }]";
        BufferedReader reader = new BufferedReader(new StringReader(requestRootWithoutAttributes));
        when(this.httpServletRequest.getReader()).thenReturn(reader);

        try{
            // Act
            sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

            // Asset
            fail("Should Throw Exception");
        } catch (AppException appException){
            Assert.assertEquals(HttpStatus.BAD_REQUEST.value(), appException.getError().getCode());
            Assert.assertEquals("Unable to parse request payload.", appException.getError().getMessage());
        } catch (Exception exception) {
            fail("Should Throw AppException");
        }
    }

    @Test
    public void should_throwWhenDataFiledIsMissing_extractDataFromRequestBody() throws IOException {
        String requestRootWithoutData = "[{\n" +
                "        \"id\": \"2425\",\n" +
                "        \"eventType\": \"recordInserted\",\n" +
                "        \"subject\": \"myapp/vehicles/motorcycles\",\n" +
                "        \"data\": {\n" +
                "            \"attributes\": {\n" +
                "                \"correlation-id\": \"39137f49-67d6-4001-a6aa-15521ef4f49e\",\n" +
                "                \"data-partition-id\": \"opendes \"\n" +
                "            },\n" +
                "            \"messageId\": \"136969346945\"\n" +
                "        },\n" +
                "        \"dataVersion\": \"1.0\",\n" +
                "        \"metadataVersion\": \"1\",\n" +
                "        \"eventTime\": \"2020-08-14T18:04:12+00:00\",\n" +
                "        \"topic\": \"/subscriptions/asdf/resourceGroups/komakkar-OSDU-RG/providers/Microsoft.EventGrid/topics/recordChanged\"\n" +
                "    }]";
        BufferedReader reader = new BufferedReader(new StringReader(requestRootWithoutData));
        when(this.httpServletRequest.getReader()).thenReturn(reader);

        try{
            // Act
            sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

            // Asset
            fail("Should Throw Exception");
        } catch (AppException appException){
            Assert.assertEquals(HttpStatus.BAD_REQUEST.value(), appException.getError().getCode());
            Assert.assertEquals("Unable to parse request payload.", appException.getError().getMessage());
        } catch (Exception exception) {
            fail("Should Throw AppException");
        }
    }

    @Test
    public void should_throwWhenPartitionIdIsMissing_extractDataFromRequestBody() throws IOException {
        String requestRootWithoutDataPartitionId = "[{\n" +
                "        \"id\": \"2425\",\n" +
                "        \"eventType\": \"recordInserted\",\n" +
                "        \"subject\": \"myapp/vehicles/motorcycles\",\n" +
                "        \"data\": {\n" +
                "            \"attributes\": {\n" +
                "                \"correlation-id\": \"39137f49-67d6-4001-a6aa-15521ef4f49e\"\n" +
                "            },\n" +
                "            \"data\": \"dGVzdA==\",\n"+
                "            \"messageId\": \"136969346945\"\n" +
                "        },\n" +
                "        \"dataVersion\": \"1.0\",\n" +
                "        \"metadataVersion\": \"1\",\n" +
                "        \"eventTime\": \"2020-08-14T18:04:12+00:00\",\n" +
                "        \"topic\": \"/subscriptions/asdf/resourceGroups/komakkar-OSDU-RG/providers/Microsoft.EventGrid/topics/recordChanged\"\n" +
                "    }]";
        BufferedReader reader = new BufferedReader(new StringReader(requestRootWithoutDataPartitionId));
        when(this.httpServletRequest.getReader()).thenReturn(reader);

        try{
            // Act
            sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

            // Asset
            fail("Should Throw Exception");
        } catch (AppException appException){
            Assert.assertEquals(HttpStatus.BAD_REQUEST.value(), appException.getError().getCode());
            Assert.assertEquals("Unable to parse request payload.", appException.getError().getMessage());
        } catch (Exception exception) {
            fail("Should Throw AppException");
        }
    }

    @Test
    public void should_returnValidData_extractDataFromRequestBody() throws IOException {
        String vaidRequestRoot = "[{\n" +
                "        \"id\": \"2425\",\n" +
                "        \"eventType\": \"recordInserted\",\n" +
                "        \"subject\": \"myapp/vehicles/motorcycles\",\n" +
                "        \"data\": {\n" +
                "            \"attributes\": {\n" +
                "                \"correlation-id\": \"39137f49-67d6-4001-a6aa-15521ef4f49e\",\n" +
                "                \"data-partition-id\": \"opendes \"\n" +
                "            },\n" +
                "            \"data\": \"dGVzdA==\",\n" +
                "            \"messageId\": \"136969346945\"\n" +
                "        },\n" +
                "        \"dataVersion\": \"1.0\",\n" +
                "        \"metadataVersion\": \"1\",\n" +
                "        \"eventTime\": \"2020-08-14T18:04:12+00:00\",\n" +
                "        \"topic\": \"/subscriptions/asdf/resourceGroups/komakkar-OSDU-RG/providers/Microsoft.EventGrid/topics/recordChanged\"\n" +
                "    }]";

        String expectedData = "test";

        BufferedReader reader = new BufferedReader(new StringReader(vaidRequestRoot));
        when(this.httpServletRequest.getReader()).thenReturn(reader);

        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

        // Act
        String receivedData = this.sut.extractDataFromRequestBody();

        // Asset
        Assert.assertEquals(expectedData,receivedData);
    }

    @Test
    public void should_returnValidAttributes_extractDataFromRequestBody() throws IOException {
        String vaidRequestRoot = "[{\n" +
                "        \"id\": \"2425\",\n" +
                "        \"eventType\": \"recordInserted\",\n" +
                "        \"subject\": \"myapp/vehicles/motorcycles\",\n" +
                "        \"data\": {\n" +
                "            \"attributes\": {\n" +
                "                \"correlation-id\": \"39137f49-67d6-4001-a6aa-15521ef4f49e\",\n" +
                "                \"data-partition-id\": \"opendes\"\n" +
                "            },\n" +
                "            \"data\": \"dGVzdA==\",\n" +
                "            \"messageId\": \"136969346945\"\n" +
                "        },\n" +
                "        \"dataVersion\": \"1.0\",\n" +
                "        \"metadataVersion\": \"1\",\n" +
                "        \"eventTime\": \"2020-08-14T18:04:12+00:00\",\n" +
                "        \"topic\": \"/subscriptions/asdf/resourceGroups/komakkar-OSDU-RG/providers/Microsoft.EventGrid/topics/recordChanged\"\n" +
                "    }]";
        BufferedReader reader = new BufferedReader(new StringReader(vaidRequestRoot));
        when(this.httpServletRequest.getReader()).thenReturn(reader);
        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

        // Act
        Map<String, String> observedAttributes = this.sut.extractAttributesFromRequestBody();

        // Asset
        Assert.assertEquals(observedAttributes.get("correlation-id"),"39137f49-67d6-4001-a6aa-15521ef4f49e");
        Assert.assertEquals(observedAttributes.get("data-partition-id"),"opendes");
    }

    @Test
    public void should_returnValidResponse_getHandshakeResponse() throws IOException {
        // Set up
        String validHandshakeRequestRoot =
                "    [{\n" +
                        "        \"id\": \"testId\",\n" +
                        "        \"topic\": \"testTopic\",\n" +
                        "        \"subject\": \"\",\n" +
                        "        \"data\": {\n" +
                        "            \"validationCode\": \"testValidationCode\",\n" +
                        "            \"validationUrl\": \"testURL\"\n" +
                        "        },\n" +
                        "        \"eventType\": \"Microsoft.EventGrid.SubscriptionValidationEvent\",\n" +
                        "        \"eventTime\": \"2020-08-14T11:18:55.9278057Z\",\n" +
                        "        \"metadataVersion\": \"1\",\n" +
                        "        \"dataVersion\": \"2\"\n" +
                        "    }]";
        String expectedResponse = "testValidationCode";
        BufferedReader reader = new BufferedReader(new StringReader(validHandshakeRequestRoot));
        when(this.httpServletRequest.getReader()).thenReturn(reader);
        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

        // Act
        String observedResponse = this.sut.getValidationCodeForHandshake();

        // Assert
        Assert.assertEquals(observedResponse, expectedResponse);
    }

    @Test
    public void should_throwWhenNotHandshakeRequest_getHandshakeResponse() throws IOException {
        // Set up
        String validHandshakeRequestRoot =
                "[{\n" +
                        "        \"id\": \"2425\",\n" +
                        "        \"eventType\": \"recordInserted\",\n" +
                        "        \"subject\": \"myapp/vehicles/motorcycles\",\n" +
                        "        \"data\": {\n" +
                        "            \"attributes\": {\n" +
                        "                \"correlation-id\": \"39137f49-67d6-4001-a6aa-15521ef4f49e\",\n" +
                        "                \"data-partition-id\": \"opendes\"\n" +
                        "            },\n" +
                        "            \"data\": \"dGVzdA==\",\n" +
                        "            \"messageId\": \"136969346945\"\n" +
                        "        },\n" +
                        "        \"dataVersion\": \"1.0\",\n" +
                        "        \"metadataVersion\": \"1\",\n" +
                        "        \"eventTime\": \"2020-08-14T18:04:12+00:00\",\n" +
                        "        \"topic\": \"/subscriptions/asdf/resourceGroups/komakkar-OSDU-RG/providers/Microsoft.EventGrid/topics/recordChanged\"\n" +
                        "    }]";
        String expectedResponse = null;
        BufferedReader reader = new BufferedReader(new StringReader(validHandshakeRequestRoot));
        when(this.httpServletRequest.getReader()).thenReturn(reader);
        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

        // Act
        String observedResponse = this.sut.getValidationCodeForHandshake();

        // Assert
        Assert.assertNull(observedResponse);
    }

    @Test
    public void should_throwWhenHandshakeRequest_extractDataFromRequestBody() throws IOException {
        String inVaidRequestRoot = "    [{\n" +
                "        \"id\": \"testId\",\n" +
                "        \"topic\": \"testTopic\",\n" +
                "        \"subject\": \"\",\n" +
                "        \"data\": {\n" +
                "            \"validationCode\": \"testValidationCode\",\n" +
                "            \"validationUrl\": \"testURL\"\n" +
                "        },\n" +
                "        \"eventType\": \"Microsoft.EventGrid.SubscriptionValidationEvent\",\n" +
                "        \"eventTime\": \"2020-08-14T11:18:55.9278057Z\",\n" +
                "        \"metadataVersion\": \"1\",\n" +
                "        \"dataVersion\": \"2\"\n" +
                "    }]";
        BufferedReader reader = new BufferedReader(new StringReader(inVaidRequestRoot));
        when(this.httpServletRequest.getReader()).thenReturn(reader);
        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

        // Act
        Map<String, String> observedAttributes = this.sut.extractAttributesFromRequestBody();

        Assert.assertNull(observedAttributes);
    }

    @Test
    public void should_returnNotificationId_extractNotificationIdFromRequestBody() throws IOException {
        // Set Up
        String vaidRequestRoot = "[{\n" +
                "        \"id\": \"2425\",\n" +
                "        \"eventType\": \"recordInserted\",\n" +
                "        \"subject\": \"myapp/vehicles/motorcycles\",\n" +
                "        \"data\": {\n" +
                "            \"attributes\": {\n" +
                "                \"correlation-id\": \"39137f49-67d6-4001-a6aa-15521ef4f49e\",\n" +
                "                \"data-partition-id\": \"opendes\"\n" +
                "            },\n" +
                "            \"data\": \"dGVzdA==\",\n" +
                "            \"messageId\": \"136969346945\"\n" +
                "        },\n" +
                "        \"dataVersion\": \"1.0\",\n" +
                "        \"metadataVersion\": \"1\",\n" +
                "        \"eventTime\": \"2020-08-14T18:04:12+00:00\",\n" +
                "        \"topic\": \"/subscriptions/asdf/resourceGroups/komakkar-OSDU-RG/providers/Microsoft.EventGrid/topics/recordChanged\"\n" +
                "    }]";
        BufferedReader reader = new BufferedReader(new StringReader(vaidRequestRoot));
        when(this.httpServletRequest.getReader()).thenReturn(reader);
        when(this.httpServletRequest.getHeader("Aeg-Subscription-Name")).thenReturn("NotificationId");
        sut = new EventGridRequestBodyExtractor(httpServletRequest, log);

        // Act
        String observed = sut.extractNotificationIdFromRequestBody();

        // Assert
        Assert.assertEquals("NotificationId", observed);

        // Set Up
        when(this.httpServletRequest.getHeader("Aeg-Subscription-Name")).thenReturn(null);

        try {
            // Act
            this.sut.extractNotificationIdFromRequestBody();

            // Asset
            fail("Should Throw Exception");
        } catch (AppException appException){
            Assert.assertEquals(HttpStatus.BAD_REQUEST.value(), appException.getError().getCode());
            Assert.assertEquals("Subscription ID not found", appException.getError().getMessage());
        } catch (Exception exception) {
            fail("Should Throw AppException");
        }
    }
}
